"""
======
Server
======

This module manages loading the modbus Server or Slave.

Examples of severs can be seen in `example/server`_

.. _example/server: https://gitlab.com/nee2c/mbsim-core/-/tree/client/examples/server

When creating a prototype.

#. Create identity (Optional)
#. Create Sever Context. Default is to full space with all values set to 0
#. Create functions or coroutine functions and add tasks to run.
#. Start the server and tasks.  See `pymodbus server
   <https://pymodbus.readthedocs.io/en/latest/source/library/server.html#pymodbus.server.StartAsyncSerialServer>`_
   to see keyword arguments to pass to the start function.
"""
import logging

from pymodbus import __version__ as version
from pymodbus.datastore import (
    ModbusServerContext,
    ModbusSlaveContext,
    ModbusSparseDataBlock,
)
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.server import (
    ServerStop,
    StartAsyncSerialServer,
    StartAsyncTcpServer,
    StartAsyncUdpServer,
)

from mbsim.core.tasks import Task, getloop

DEFAULTCONF = {
    "tcp": {"context": None, "identity": None, "address": ("", 502), "ignore_missing_slaves": False},
    "udp": {"context": None, "identity": None, "address": ("", 502), "ignore_missing_slaves": False},
    "rtu": {
        "context": None,
        "identity": None,
        "port": "/dev/ttyS0",
        "buadrate": 19200,
        "bytesize": 8,
        "stopbits": 1,
        "parity": "N",
        "timeout": 0,
        "xonxoff": 0,
        "rtscts": 0,
    },
}

log = logging.getLogger(__name__)
log.debug("pymodbus version: %s", version)


def genDevice(
    name="mbsim",
    code="MB",
    url="https://gitlab.com/nee2c/mbsim",
    product="mbsim",
    model="mbsim",
    version="1.0.0",
):
    """
    A function to create a identity.  All parameters are key word arguments.

    :param name: Vendor Name. Defaults to `mbsim`
    :type name: str
    :param code: Product Code. Defaults to `MB`
    :type code: str
    :param url: Vendor URL. Defaults to `https://gitlab.com/nee2c/mbsim`
    :type url: str
    :param product: Product Name. Defaults to `mbsim`
    :type product: str
    :param model: Model Name. Defaults to `mbsim`
    :type model: str
    :param version: Version. Defaults to `1.0.0`
    :type version: str
    :return: Returns an identity for server
    """
    identity = ModbusDeviceIdentification()

    identity.VendorName = name
    identity.ProductCode = code
    identity.VendorUrl = url
    identity.ProductName = product
    identity.ModelName = model
    identity.MajorMinorRevision = version

    log.debug("Generated identity: %s", identity.summary())
    return identity


def genContext(context=None, single=True):
    """
    A Function to return Context for Server.

    The context can be None. This will generate a context that all slavesid will use. All values set to 0

    If context is an instance of ModbusServerContext, function returns the context

    Else dict of slaves with dict of address space("di", "co", "hr", "ir") with list or dict {offset: [reg0, reg1, ...]}


    dict Context example

    .. code::

        {0: {"di": [0, 1, 0, 1, 0], "co": {0: [1, 0, 1], 100: [1]}, "hr": [123, 10], "ir": {123: [123, 0, 123]}}, ...}

    :param context: Server Context or dict of Slaves context or dict
    :param single: This makes all slave id maps to one Slave context
    :type single: bool
    """
    if isinstance(context, ModbusServerContext):
        log.debug("Already a Server Context")
        return context
    if context is None:
        log.debug("Generating Server context: single: %s", single)
        return ModbusServerContext(slaves=ModbusSlaveContext(zero_mode=True), single=single)
    slaves = {}
    log.debug("Generating context from %s", context)
    for slaveid, slavecontext in context.items():
        slaves[slaveid] = ModbusSlaveContext(
            **{key: ModbusSparseDataBlock(values=vals) for key, vals in slavecontext.items()},
            zero_mode=True,
        )
    return ModbusServerContext(slaves=slaves, single=single)


def start(server, loop=None, **kwargs):
    """
    This is the function to start modbus server.  Passes all kwargs though to server, if missing uses default.

    To see full list of keyword arguments for supported protocols see.

    - `rtu`_
    - `tcp`_
    - `udp`_

    .. _rtu: https://pymodbus.readthedocs.io/en/latest/source/library/server.html#pymodbus.server.StartAsyncSerialServer
    .. _tcp: https://pymodbus.readthedocs.io/en/latest/source/library/server.html#pymodbus.server.StartAsyncTcpServer
    .. _udp: https://pymodbus.readthedocs.io/en/latest/source/library/server.html#pymodbus.server.StartAsyncUdpServer

    :param server: The protocol to run modbus server.
    :type server: str
    :param loop: The event loop and if none will use running loop or create new loop
    """
    if server not in DEFAULTCONF.keys():
        raise NotImplementedError("This server type {} is not Implemented".format(server))
    loop = loop or getloop()
    serverargs = {**kwargs, **{key: val for key, val in DEFAULTCONF[server].items() if key not in kwargs.keys()}}
    if not serverargs.get("context"):
        serverargs["context"] = genContext()
    log.info("Starting %s server", server)
    log.debug("Server args: %s", str(serverargs))

    log.debug("Starting Tasks")
    Task.startTasks(loop=loop)
    log.debug("Started Tasks")

    log.debug("Starting %s Server", server)
    if server == "tcp":
        loop.create_task(StartAsyncTcpServer(**serverargs))
    elif server == "udp":
        loop.create_task(StartAsyncUdpServer(**serverargs))
    elif server == "rtu":  # pragma: no cover  # there is a precondition check
        loop.create_task(StartAsyncSerialServer(**serverargs))
    log.debug("Started Server")
    loop.run_forever()
    loop.close()


def stop():
    """
    Function to stop modbus server
    """
    log.debug("Stopping Server")
    ServerStop()
    log.debug("Stopped Server")
    log.debug("Stopping Tasks")
    Task.stopTasks()
    log.debug("Stopped Tasks")



"""
======
Client
======

This module manages loading the modbus Client or Master.

Examples of severs can be seen in `example/client`_

.. _example/client: https://gitlab.com/nee2c/mbsim-core/-/tree/client/examples/client

To start prototyping

#. Create clients
#. Create Tasks
#. Start Tasks

The Clients that we suggest using can be found `here`_

.. _here: https://pymodbus.readthedocs.io/en/latest/source/library/client.html#
"""
import logging

from pymodbus.client import (  # noqa: F401
    AsyncModbusSerialClient,
    AsyncModbusTcpClient,
    AsyncModbusUdpClient,
    ModbusSerialClient,
    ModbusTcpClient,
    ModbusUdpClient,
)

from mbsim.core.tasks import Task, getloop  # noqa: F401

log = logging.getLogger(__name__)


def start(loop=None):
    """
    This is the function to start Tasks for client

    :param loop: The event loop and if none will create new loop or use running loop
    """
    if loop is None:
        loop = getloop()
    log.debug("Starting Tasks")
    Task.startTasks(loop=loop)
    log.debug("Started Tasks")
    loop.run_forever()
    loop.close()


def stop():
    """
    Function to stop Tacks
    """
    log.debug("Stopping Tasks")
    Task.stopTasks()
    log.debug("Stopped Tasks")



"""
====
Task
====

A module to manage classes of Task objects.

Task are a way to run a function or coroutinefunction periodically.  This is useful for things like updating a values
or sending a message to a client every so often.

.. code::

    @mbsim.core.server.Task(1, args=("Hello", "world"))
    def hw(*args):
        print(*args)

    def main():
        task = mbsim.core.server.Task(1, func=hw, args=("foo", "bar"))
        mb.start()


This will print "Hello" and "world" every second. I also prints "foo" and "bar" every second.
"""
import asyncio
import inspect
import logging
import warnings
from functools import partial

log = logging.getLogger(__name__)


def getloop():
    """
    Function to return async loop
    """
    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.new_event_loop()


class Task(object):
    """
    A Class to manage all the task to run periodically alongside the server/client

    This can be called as a decorators.  The following example will print "Hello\\nworld" every second.
    The decorator can be used multiple times to call more than once.

    .. code::

        @mbsim.core.server.Task(1, args=("Hello", "world"))
        def hw(*args):
            for arg in args:
                print(arg)

        or

        mbsim.core.server.Task(1, func=hw, args=("Hello", "world"))

    Both example are equivalent
    """

    tasks = []
    loop = None
    _clean = None

    def __init__(self, inter, func=None, args=(), kwargs=None, now=False, block=False):
        """
        This will take the arguments for the task to be run

        :param inter: Interval of time to wait before calling the function or coroutinefunction in seconds
        :type inter: float
        :param func: Function to call
        :type func: function, coroutinefunction, optional
        :param args: arguments to be past to task
        :type args: tuple, optional
        :param kwargs: key word arguments to be passed to task, defaults to {}
        :type kwargs: dict, optional
        :param now: To run immediately or wait the interval, defaults to False
        :type now: bool, optional
        """
        if kwargs is None:
            kwargs = {}
        self.task = None
        self.inter = inter
        self.args = args
        self.kwargs = kwargs
        self.now = now
        self.tasks.append(self)
        self.block = block
        if func:
            self(func)

        warnings.warn("The Block argument will be removed in v1.1.0", DeprecationWarning)

    def __call__(self, func):
        """
        Init the task and return original function

        :param func: A Non blocking function
        :type func: function
        :return: Original function
        :rtype: function
        """
        self.func = func
        return func

    async def wrap(self, a, kw):
        """
        A wrap for twisted to pass a and kw as args.  This function will explode them into the function

        :param a: args
        :type a: tuple
        :param kw: kwargs
        :type kw: dict
        """
        try:
            if inspect.iscoroutinefunction(self.func):
                if self.block:
                    raise ValueError("{} cannot be blocking if is coroutine".format(self.func.__name__))

                if self.now:
                    await self.func(*a, **kw)

                while True:
                    await asyncio.sleep(self.inter)
                    await self.func(*a, **kw)

            if self.now:
                await self.loop.run_in_executor(None, partial(self.func, *a, **kw))

            while True:
                await asyncio.sleep(self.inter)
                await self.loop.run_in_executor(None, partial(self.func, *a, **kw))

        except asyncio.CancelledError:
            log.debug("Task %s canceled", (self.func.__name__, self.inter, self.args, self.kwargs))

    def start(self):
        """
        To start the task
        """
        log.debug("Starting task: %s", (self.func.__name__, self.inter, self.args, self.kwargs))
        self.task = self.loop.create_task(self.wrap(self.args, self.kwargs))

    @classmethod
    def startTasks(cls, loop=None):
        """
        Class method to start all tasks
        """
        if cls.loop and loop:
            raise RuntimeError("Cannot replace loop in Tasks")
        elif cls.loop is None:
            if loop is not None:
                cls.loop = loop
            else:
                cls.loop = getloop()

        cls._clean = cls.loop.create_task(_cleanup(cls.loop))

        for task in cls.tasks:
            task.start()

    def stop(self):
        """
        To stop a task
        """
        log.debug("Stopping task: %s", (self.func.__name__, self.inter, self.args, self.kwargs))
        self.task.cancel()

    @classmethod
    def stopTasks(cls):
        """
        Class method to stop all tasks
        """
        for task in cls.tasks:
            task.stop()

        cls._clean.cancel()


async def _cleanup(loop):
    """
    Coroutine to allow all async tasks to clean up

    :param loop: Event loop
    """
    try:
        while True:
            await asyncio.sleep(0.05)
    except asyncio.CancelledError:
        await asyncio.sleep(1)
        loop.stop()



"""
Test for tasks module
"""
import asyncio

import pytest
from mock import MockCall, MockLoopingCall

from mbsim.core.tasks import Task, _cleanup, getloop

pytest_plugins = ("pytest_asyncio",)


class TestGetLoop(object):
    """
    This test the getloop function
    """

    def raiseRintime(self):
        raise RuntimeError

    def test_new_event_loop(self, monkeypatch):
        """
        Test if new event loop is passed
        """
        monkeypatch.setattr(asyncio, "get_running_loop", self.raiseRintime)
        loop = getloop()
        assert isinstance(loop, asyncio.BaseEventLoop)

    @pytest.mark.asyncio
    async def test_running_loop(self, event_loop):
        """
        Test running loop is passed
        """
        loop = getloop()
        assert loop is event_loop


class Test_Tasks(object):
    """
    Group of test to test Tasks
    """

    @pytest.mark.parametrize(
        "inter,args,kwargs,now,func",
        [
            (1, (), {}, False, print),
            (2, ("Hello"), {}, False, dict),
            (3, (), {"Hello": "World"}, False, set),
            (4, (), None, True, max),
        ],
    )
    def test_createDecorator(self, inter, args, kwargs, now, func, monkeypatch):
        """
        test the creation of tasks
        """
        Task(inter, args=args, kwargs=kwargs, now=now)(func)
        assert func is Task.tasks[-1].func
        assert inter is Task.tasks[-1].inter
        assert args is Task.tasks[-1].args
        assert now is Task.tasks[-1].now
        if kwargs is None:
            assert isinstance(Task.tasks[-1].kwargs, dict)
        else:
            assert kwargs is Task.tasks[-1].kwargs

    @pytest.mark.parametrize(
        "inter,args,kwargs,now,func",
        [
            (1, (), {}, False, print),
            (2, ("Hello"), {}, False, dict),
            (3, (), {"Hello": "World"}, False, set),
            (4, (), None, True, max),
        ],
    )
    def test_createInstance(self, inter, args, kwargs, now, func, monkeypatch):
        """
        test the creation of tasks
        """
        Task(inter, func=func, args=args, kwargs=kwargs, now=now)
        assert func is Task.tasks[-1].func
        assert inter is Task.tasks[-1].inter
        assert args is Task.tasks[-1].args
        assert now is Task.tasks[-1].now
        if kwargs is None:
            assert isinstance(Task.tasks[-1].kwargs, dict)
        else:
            assert kwargs is Task.tasks[-1].kwargs

    @pytest.mark.parametrize(
        "tasks",
        [[], [MockLoopingCall(lambda: None)], [MockLoopingCall(lambda: None) for _ in range(3)]],
    )
    def test_startTasks(self, tasks, monkeypatch):
        """
        Test class method and by proxy instance start method
        """
        monkeypatch.setattr(Task, "tasks", tasks)
        Task.startTasks()
        for task in Task.tasks:
            assert task.started

    @pytest.mark.parametrize(
        "tasks",
        [[], [MockLoopingCall(lambda: None)], [MockLoopingCall(lambda: None) for _ in range(3)]],
    )
    def test_stopTasks(self, tasks, monkeypatch):
        """
        Test class method and by proxy instance stop method
        """
        monkeypatch.setattr(Task, "tasks", tasks)
        Task.stopTasks()
        for task in Task.tasks:
            assert task.stopped

    def test_startTask(self, monkeypatch):
        """
        Test starting a task
        """
        loop = MockLoopingCall(lambda: None)
        monkeypatch.setattr(Task, "loop", loop)
        task = Task(1, func=print)
        monkeypatch.setattr(task, "wrap", MockCall())
        task.start()
        assert loop.task

    def test_stopTask(self, monkeypatch):
        """
        Test stopping a task
        """
        mtask = MockLoopingCall(lambda: None)
        task = Task(1, func=print)
        monkeypatch.setattr(task, "task", mtask)
        task.stop()
        assert mtask.stopped

    @pytest.mark.parametrize("block", [True, False])
    @pytest.mark.parametrize("now", [True, False])
    @pytest.mark.asyncio
    async def test_funcwrap(self, event_loop, now, block, monkeypatch):
        """
        Test wrap function for async Call passing args
        """
        mc = MockCall()
        monkeypatch.setattr(Task, "loop", event_loop)
        task = Task(0.3, func=mc, now=now, block=block)
        test = event_loop.create_task(task.wrap(a=("hello",), kw={"a": 1}))
        await asyncio.sleep(0.4)
        test.cancel()
        if now:
            assert mc.count == 2
        else:
            assert mc.count == 1
        for val in mc.args:
            assert val[0] == "hello"
        for val in mc.kwargs:
            assert val == {"a": 1}

    @pytest.mark.asyncio
    async def test_blockcoroutine(self, event_loop, monkeypatch):
        """
        Test that giving task a coroutine and set blocking true
        """
        with pytest.raises(ValueError, match="{} cannot be blocking if is coroutine".format(asyncio.sleep.__name__)):
            monkeypatch.setattr(Task, "loop", event_loop)
            task = Task(1, func=asyncio.sleep, args=(1,), block=True)
            task.start()
            await task.task

    @pytest.mark.parametrize("now", (True, False))
    @pytest.mark.asyncio
    async def test_asyncwrap(self, event_loop, now, monkeypatch):
        """
        Test if a coroutine is passed to task
        """

        async def test(mc):
            """Mock coroutine"""
            mc()

        mc = MockCall()
        monkeypatch.setattr(Task, "loop", event_loop)
        task = Task(0.3, func=test, now=now)
        test = event_loop.create_task(task.wrap(a=(mc,), kw={}))
        await asyncio.sleep(0.4)
        if now:
            assert mc.count == 2
        else:
            assert mc.count == 1

    def test_reaisedoubeloop(self, event_loop):
        """
        Test to makesure if Task raise exception if loop is already defined
        """
        with pytest.raises(RuntimeError, match="Cannot replace loop in Tasks"):
            Task.loop = event_loop
            Task.startTasks(loop=event_loop)


@pytest.mark.asyncio
async def test__cleanup(event_loop):
    """
    Test if as cleanup functions works
    """
    mock = MockLoopingCall(lambda: None)
    task = event_loop.create_task(_cleanup(mock))
    await asyncio.sleep(0.1)
    task.cancel()
    await asyncio.sleep(1.1)
    assert task.done()
    assert mock.stopped



"""
Test for mbsim-core.server module
"""
from importlib import import_module, reload
from itertools import combinations

import pytest
from mock import MockCall, MockLoopingCall
from pymodbus.datastore import ModbusServerContext, ModbusSlaveContext

import mbsim.core.server as mbserver

pytest_plugins = ("pytest_asyncio",)


@pytest.fixture(scope="function")
def mb():
    """
    Import lib for testing and reload
    """
    mb = import_module("mbsim.core.server")
    yield mb
    reload(mb)
    mb.Task.loop = None
    mb.Task.tasks = []


def genDevData():
    """
    Genrates device identity test data
    """
    default = {
        "VendorName": "mbsim",
        "ProductCode": "MB",
        "VendorUrl": "https://gitlab.com/nee2c/mbsim",
        "ProductName": "mbsim",
        "ModelName": "mbsim",
        "MajorMinorRevision": "1.0.0",
    }
    alt = {
        "VendorName": "tests",
        "ProductCode": "AZ",
        "VendorUrl": "https://example.com",
        "ProductName": "sim",
        "ModelName": "mb",
        "MajorMinorRevision": "1.0.1",
    }
    paramname = {
        "VendorName": "name",
        "ProductCode": "code",
        "VendorUrl": "url",
        "ProductName": "product",
        "ModelName": "model",
        "MajorMinorRevision": "version",
    }
    data = [(None, default)]
    for num, _ in enumerate(default, 1):
        for keys in combinations(default.keys(), num):
            test = {paramname[key]: alt[key] for key in keys}
            res = {
                **{key: alt[key] for key in keys},
                **{key: default[key] for key in set(default.keys()).difference(keys)},
            }
            data.append((test, res))
    return data


def genContext():
    return mbserver.genContext()


class TestGenDevice(object):
    """
    This is group test for the function of genDevice
    """

    @pytest.mark.parametrize("test,res", genDevData())
    def test_genDevice(self, test, res, mb):
        """
        Test identity after it genrated
        """
        if test is None:
            test = {}
        dev = mb.genDevice(**test)
        for key, value in res.items():
            assert getattr(dev, key) == value


class TestGenContext(object):
    """
    This is a group of text for testing generating Context
    """

    _fx = {2: "d", 4: "i", **{i: "h" for i in (3, 6, 16, 22, 23)}, **{i: "c" for i in (1, 5, 15)}}

    def test_passingNone(self, mb):
        """
        test calling genContext with no arguments
        """
        test = mb.genContext()
        assert isinstance(test, ModbusServerContext)
        assert test.single
        assert test[0] is test[1]
        for fx in self._fx.keys():
            assert test[0].getValues(fx, 200, count=5)

    def test_passingContext(self, mb):
        """
        test passing the function exsisting Server Context
        """
        test = ModbusServerContext()
        assert mb.genContext(test) is test

    @pytest.mark.parametrize(
        "testdict",
        [
            (
                {
                    2: {
                        "co": {0: [1, 0, 0, 1], 101: [0, 1, 1, 0]},
                        "di": {0: [1, 0, 0, 1], 201: [0, 1, 1, 0]},
                        "hr": {60: [100, 580, 40, 1], 101: [0, 101, 1111, 0]},
                        "ir": {10: [1000, 0, 0, 541], 101: [870, 221, 1, 0]},
                    }
                }
            ),
            (
                {
                    2: {
                        "co": [1, 0, 0, 1],
                        "di": [1, 0, 0, 1],
                        "hr": [100, 580, 40, 1],
                        "ir": [1000, 0, 0, 5410, 221, 1, 0],
                    }
                }
            ),
            (
                {
                    2: {
                        "co": [1, 0, 0, 1],
                        "di": [1, 0, 0, 1],
                        "hr": {60: [100, 580, 40, 1], 101: [0, 101, 1111, 0]},
                        "ir": {10: [1000, 0, 0, 541], 101: [870, 221, 1, 0]},
                    }
                }
            ),
            (
                {
                    2: {
                        "co": [1, 0, 0, 1],
                        "di": [1, 0, 0, 1],
                        "hr": [100, 580, 40, 1],
                        "ir": [1000, 0, 0, 5410, 221, 1, 0],
                    },
                    20: {
                        "co": [1, 1, 1, 1],
                        "di": [0, 0, 0, 0],
                        "hr": [107, 587, 44, 6],
                        "ir": [1006, 4, 0, 5402, 221, 4, 0],
                    },
                }
            ),
        ],
    )
    def test_customContext(self, testdict, mb):
        """
        create and test custom setup context
        """
        _fx = {"co": 1, "di": 2, "hr": 3, "ir": 4}
        test = mb.genContext(testdict, single=False)
        assert isinstance(test, ModbusServerContext)
        for slave in test.slaves():
            assert slave in testdict
            assert isinstance(test[slave], ModbusSlaveContext)
            for store, vals in testdict[slave].items():
                if isinstance(vals, dict):
                    for offset, regs in vals.items():
                        assert test[slave].getValues(_fx[store], offset, len(regs)) == regs
                elif isinstance(vals, list):
                    assert test[slave].getValues(_fx[store], 0, len(vals)) == vals


class Test_Start(object):
    """
    Group of test to test the start function
    """

    def test_notimplemented(self, mb):
        """
        Test if unexpected server raise NotImplementedError
        """
        with pytest.raises(NotImplementedError, match="This server type Hello is not Implemented"):
            mb.start("Hello")

    @pytest.mark.parametrize("context", [None, genContext()])
    def test_Context(self, context, monkeypatch, mb):
        """
        Test to check behavior with different types of context is used as prams
        """
        server, mock, loop = "tcp", MockCall(), MockLoopingCall(lambda: None)
        monkeypatch.setattr(mb, "StartAsyncTcpServer", mock)
        monkeypatch.setattr(mb, "getloop", lambda: loop)
        if context is None:
            # Call with no context param
            mb.start(server)
        else:
            mb.start(server, context=context)

        assert loop.task == 1
        assert loop.forever
        assert isinstance(mock.kwargs[0].get("context"), ModbusServerContext)

    @pytest.mark.parametrize("test", ["tcp", "udp", "rtu"])
    def test_servercalls(self, test, monkeypatch, mb):
        """
        Test the pymodbus module is called correctly
        """
        servers = {
            "tcp": ("StartAsyncTcpServer", MockCall()),
            "udp": ("StartAsyncUdpServer", MockCall()),
            "rtu": ("StartAsyncSerialServer", MockCall()),
        }
        loop = MockLoopingCall(lambda: None)
        monkeypatch.setattr(mb, "getloop", lambda: loop)
        for server, (attr, mock) in servers.items():
            monkeypatch.setattr(mb, attr, mock)
        mb.start(
            test,
        )

        assert loop.task == 1
        assert loop.forever
        _, res = servers.pop(test)
        assert res.count == 1
        for _, mock in servers.values():
            assert not mock.count

    @pytest.mark.parametrize("loop", [None, MockLoopingCall(lambda: None)])
    def test_startTasks(self, mb, monkeypatch, loop):
        """
        Test to make sure Tasks are started
        """
        test = MockCall()
        mloop = loop or MockLoopingCall(lambda: None)
        monkeypatch.setattr(mb, "getloop", lambda: mloop)
        monkeypatch.setattr(mb, "StartAsyncTcpServer", MockCall())
        monkeypatch.setattr(mb.Task, "startTasks", test)
        mb.start("tcp", loop=loop)
        assert test.count == 1


class TestStop(object):
    """
    Test Stop functions
    """

    def test_stop(self, monkeypatch, mb):
        """
        Test Stop functions
        """
        mock = MockCall()
        mock2 = MockCall()
        monkeypatch.setattr(mb, "ServerStop", mock)
        monkeypatch.setattr(mb.Task, "stopTasks", mock2)
        mb.stop()
        assert mock.count == 1
        assert not mock.args[0]
        assert not mock.kwargs[0]
        assert mock2.count == 1
        assert not mock2.args[0]
        assert not mock2.kwargs[0]

    def test_stopTasks(self, mb, monkeypatch):
        """
        Test to make sure Tasks are started
        """
        test = MockCall()
        loop = MockLoopingCall(lambda: None)
        monkeypatch.setattr(mb, "ServerStop", MockCall())
        monkeypatch.setattr(mb.Task, "stopTasks", test)
        monkeypatch.setattr(mb.Task, "loop", loop)
        mb.stop()
        assert test.count == 1



"""
Module with a few mock objects for testing
"""


class MockCall(object):
    """
    A simple Test function to count the number of calls and arguments used
    """

    def __init__(self):
        """
        Init the test function
        """
        self.__name__ = "mock"
        self.count = 0
        self.args = []
        self.kwargs = []

    def __call__(self, *args, **kwargs):
        """
        Increment the count due to being called
        """
        self.count += 1
        self.args.append(args)
        self.kwargs.append(kwargs)


class MockLoopingCall(object):
    """
    Mock object to mock twisted LoopingCall object
    """

    def __init__(self, f, a=None, kw=None):
        """
        test object to test LoopingCall

        :param f: function
        :param a: args, defaults to None
        :type a: tuple, optional
        :param kw: kwargs, defaults to None
        :type kw: dict, optional
        """
        self.started = False
        self.tasks = 0
        self.stopped = False
        self.func = f
        self.args = a
        self.kwargs = kw
        self.inter = None
        self.now = None
        self.forever = False

    def create_task(self, inter, now=False):
        """
        mock loopingCall start function

        :param inter: time interval
        :type inter: float
        :param now: first call or wait interval, defaults to False
        :type now: bool, optional
        """
        self.task = +1
        self.inter = inter
        self.now = now
        return self

    def cancel(self):
        """
        mock loopingCall stop function
        """
        if self.stopped:
            raise RuntimeError("Mock Task has already been stopped")
        self.stopped = True

    def stop(self):
        """
        alias for cancel
        """
        self.cancel()

    def close(self):
        """
        alias for cancel
        """
        self.cancel()

    def start(self):
        """
        alias for create task
        """
        if self.started:
            raise RuntimeError("Mock Task has already been started")
        self.started = True

    def run_forever(self):
        """
        Check to see if run forever is called
        """
        self.forever = True



import mbsim.core.client as mb

client = mb.ModbusTcpClient("127.0.0.1", 5020)


@mb.Task(1, args=(client, 1, 2))
def add(client, reg, val, slave=0):
    """
    function to read holding register and add a value to it

    :param client: The modbus client
    :param reg: the holding register to read and write too
    :type reg: int
    :param val: the value to increase the register by
    :type val: int
    :param slave: Slave id
    :type slave: int
    """
    if not client.connected:
        client.connect()

    read = client.read_holding_registers(reg, 1, slave=slave)
    client.write_register(reg, (read.registers[0] + val) % 2**16, slave=slave)
    print(read.registers[0])


if __name__ == "__main__":
    mb.start()



import mbsim.core.client as mb


async def poll(client, reg, count=1, slave=0):
    """
    A coroutine function to query the values in register and prent to output

    :param client: The modbus client
    :param reg: the registers to start to read from
    :type reg: int
    :param count: Number of registers to read
    :type count: int
    :param slave: the slave id
    :type slave: int
    """
    if not client.connected:
        await client.connect()

    read = await client.read_holding_registers(reg, count, slave)
    print("Vals starting at {} reg: {}".format(reg, read.registers))


if __name__ == "__main__":
    mb.Task(1, func=poll, args=(mb.AsyncModbusTcpClient("127.0.0.1", 5020), 0, 5))
    mb.start()



import mbsim.core.server as mb

_fx = {"d": 2, "i": 4, "h": 3, "c": 1}  # Map to some modbus functions
context = mb.genContext()  # Gen a shared context with all Vals set to 0


@mb.Task(1, args=(context,))
def watchdog(context):
    """
    A Simple watch dog that will toggle the val in the first Holding register.
    The one in the decorator is the number of secs to wait till next execution.

    :param context: context that the modbus server will use
    """
    if context[0].getValues(_fx["h"], 1)[0] == 0:
        context[0].setValues(_fx["h"], 1, [1])
    else:
        context[0].setValues(_fx["h"], 1, [0])
    add(context, 4)  # Cam still call functions that have Task decorator like normal


@mb.Task(2, args=(context, 3))  # Can have multiple Tasks decorators
@mb.Task(1, args=(context, 2))
def add(context, reg):
    """
    A Simple function that will add 1 to the val in the Second Holding register
    The one in the decorator is the number of secs to wait till next execution.

    :param context: context that the modbus server will use
    """
    x = context[0].getValues(_fx["h"], reg)[0]
    x += 1
    if 0 <= x < 65536:
        context[0].setValues(_fx["h"], reg, [x])
    else:
        context[0].setValues(_fx["h"], reg, [0])


mb.start("tcp", context=context, address=("", 5020))  # Starts the modbus server



"""
Test for client module
"""
import pytest
from mock import MockCall, MockLoopingCall

import mbsim.core.client as mb

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.parametrize("loop", [None, MockLoopingCall(lambda: None)])
def testStart(monkeypatch, loop):
    """
    Test client Start function
    """
    genloop = loop or MockLoopingCall(lambda: None)
    starttasks = MockCall()

    monkeypatch.setattr(mb, "getloop", lambda: genloop)
    monkeypatch.setattr(mb.Task, "startTasks", starttasks)

    mb.start(loop=loop)

    assert starttasks.count == 1
    assert starttasks.kwargs[0]["loop"] is genloop
    assert genloop.forever


def testStop(monkeypatch):
    """
    Test client Stop function
    """
    mc = MockCall()
    monkeypatch.setattr(mb.Task, "stopTasks", mc)

    mb.stop()

    assert mc.count == 1
